@@ -36,6 +36,7 @@
     * :meth:`AdbDevice._pull`
     * :meth:`AdbDevice._push`
     * :meth:`AdbDevice._read`
+    * :meth:`AdbDevice._read_length`
     * :meth:`AdbDevice._read_until`
     * :meth:`AdbDevice._read_until_close`
     * :meth:`AdbDevice._send`
@@ -89,6 +90,8 @@ class AdbDevice(object):
     ----------
     transport : BaseTransport
         A user-provided transport for communicating with the device; must be an instance of a subclass of :class:`~adb_shell.transport.base_transport.BaseTransport`
+    default_transport_timeout_s : float, None
+        Default timeout in seconds for transport packets, or ``None``
     banner : str, bytes, None
         The hostname of the machine where the Python interpreter is currently running; if
         it is not provided, it will be determined via ``socket.gethostname()``
@@ -104,6 +107,8 @@ class AdbDevice(object):
         Whether an ADB connection to the device has been established
     _banner : bytearray, bytes
         The hostname of the machine where the Python interpreter is currently running
+    _default_transport_timeout_s : float, None
+        Default timeout in seconds for transport packets, or ``None``
     _maxdata: int
         Maximum amount of data in an ADB packet
     _transport : BaseTransport
@@ -111,7 +116,7 @@ class AdbDevice(object):
 
     """
 
-    def __init__(self, transport, banner=None):
+    def __init__(self, transport, default_transport_timeout_s=None, banner=None):
         if banner and not isinstance(banner, (bytes, bytearray)):
             self._banner = bytearray(banner, 'utf-8')
         else:
@@ -123,8 +128,26 @@ class AdbDevice(object):
         self._transport = transport
 
         self._available = False
+        self._default_transport_timeout_s = default_transport_timeout_s
         self._maxdata = constants.MAX_PUSH_DATA
 
+    # ======================================================================= #
+    #                                                                         #
+    #                       Properties & simple methods                       #
+    #                                                                         #
+    # ======================================================================= #
+    @property
+    def available(self):
+        """Whether or not an ADB connection to the device has been established.
+
+        Returns
+        -------
+        bool
+            ``self._available``
+
+        """
+        return self._available
+
     @property
     def max_chunk_size(self):
         """Maximum chunk size for filesync operations
@@ -137,18 +160,27 @@ class AdbDevice(object):
         """
         return min(constants.MAX_CHUNK_SIZE, self._maxdata // 2) or constants.MAX_PUSH_DATA
 
-    @property
-    def available(self):
-        """Whether or not an ADB connection to the device has been established.
+    def _get_transport_timeout_s(self, transport_timeout_s):
+        """Use the provided ``transport_timeout_s`` if it is not ``None``; otherwise, use ``self._default_transport_timeout_s``
+
+        Parameters
+        ----------
+        transport_timeout_s : float, None
+            The potential transport timeout
 
         Returns
         -------
-        bool
-            ``self._available``
+        float
+            ``transport_timeout_s`` if it is not ``None``; otherwise, ``self._default_transport_timeout_s``
 
         """
-        return self._available
+        return transport_timeout_s if transport_timeout_s is not None else self._default_transport_timeout_s
 
+    # ======================================================================= #
+    #                                                                         #
+    #                             Close & Connect                             #
+    #                                                                         #
+    # ======================================================================= #
     def close(self):
         """Close the connection via the provided transport's ``close()`` method.
 
@@ -212,7 +244,7 @@ class AdbDevice(object):
 
         # 2. Send a ``b'CNXN'`` message
         msg = AdbMessage(constants.CNXN, constants.VERSION, constants.MAX_ADB_DATA, b'host::%s\0' % self._banner)
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+        adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s)
         self._send(msg, adb_info)
 
         # 3. Unpack the ``cmd``, ``arg0``, ``arg1``, and ``banner`` fields from the response
@@ -297,7 +329,7 @@ class AdbDevice(object):
             The output of the ADB command as a string if ``decode`` is True, otherwise as bytes.
 
         """
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s, timeout_s)
+        adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s, timeout_s)
         if decode:
             return b''.join(self._streaming_command(service, command, adb_info)).decode('utf8')
         return b''.join(self._streaming_command(service, command, adb_info))
@@ -325,7 +357,7 @@ class AdbDevice(object):
             The line-by-line output of the ADB command as a string if ``decode`` is True, otherwise as bytes.
 
         """
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+        adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s)
         stream = self._streaming_command(service, command, adb_info)
         if decode:
             for line in (stream_line.decode('utf8') for stream_line in stream):
@@ -334,6 +366,36 @@ class AdbDevice(object):
             for line in stream:
                 yield line
 
+    def exec_out(self, command, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S, timeout_s=None, decode=True):
+        """Send an ADB ``exec-out`` command to the device.
+
+        https://www.linux-magazine.com/Issues/2017/195/Ask-Klaus
+
+        Parameters
+        ----------
+        command : str
+            The exec-out command that will be sent
+        transport_timeout_s : float, None
+            Timeout in seconds for sending and receiving packets, or ``None``; see :meth:`BaseTransport.bulk_read() <adb_shell.transport.base_transport.BaseTransport.bulk_read>`
+            and :meth:`BaseTransport.bulk_write() <adb_shell.transport.base_transport.BaseTransport.bulk_write>`
+        read_timeout_s : float
+            The total time in seconds to wait for a ``b'CLSE'`` or ``b'OKAY'`` command in :meth:`AdbDevice._read`
+        timeout_s : float, None
+            The total time in seconds to wait for the ADB command to finish
+        decode : bool
+            Whether to decode the output to utf8 before returning
+
+        Returns
+        -------
+        bytes, str
+            The output of the ADB exec-out command as a string if ``decode`` is True, otherwise as bytes.
+
+        """
+        if not self.available:
+            raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDevice.connect()`?)")
+
+        return self._service(b'exec', command.encode('utf8'), transport_timeout_s, read_timeout_s, timeout_s, decode)
+
     def root(self, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S, timeout_s=None):
         """Gain root access.
 
@@ -433,10 +495,12 @@ class AdbDevice(object):
             Filename, mode, size, and mtime info for the files in the directory
 
         """
+        if not device_path:
+            raise exceptions.DevicePathInvalidError("Cannot list an empty device path")
         if not self.available:
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDevice.connect()`?)")
 
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+        adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s)
         filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_LIST_FORMAT, maxdata=self._maxdata)
         self._open(b'sync:', adb_info)
 
@@ -471,16 +535,20 @@ class AdbDevice(object):
             The total time in seconds to wait for a ``b'CLSE'`` or ``b'OKAY'`` command in :meth:`AdbDevice._read`
 
         """
+        if not device_path:
+            raise exceptions.DevicePathInvalidError("Cannot pull from an empty device path")
         if not self.available:
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDevice.connect()`?)")
 
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+        adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s)
         filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_PULL_FORMAT, maxdata=self._maxdata)
 
         with open(local_path, 'wb') as stream:
             self._open(b'sync:', adb_info)
-            self._pull(device_path, stream, progress_callback, adb_info, filesync_info)
-            self._close(adb_info)
+            try:
+                self._pull(device_path, stream, progress_callback, adb_info, filesync_info)
+            finally:
+                self._close(adb_info)
 
     def _pull(self, device_path, stream, progress_callback, adb_info, filesync_info):
         """Pull a file from the device into the file-like ``local_path``.
@@ -534,6 +602,8 @@ class AdbDevice(object):
             The total time in seconds to wait for a ``b'CLSE'`` or ``b'OKAY'`` command in :meth:`AdbDevice._read`
 
         """
+        if not device_path:
+            raise exceptions.DevicePathInvalidError("Cannot push to an empty device path")
         if not self.available:
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDevice.connect()`?)")
 
@@ -543,7 +613,7 @@ class AdbDevice(object):
             self.shell("mkdir " + device_path, transport_timeout_s, read_timeout_s)
 
         for _local_path, _device_path in zip(local_paths, device_paths):
-            adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+            adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s)
             filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_PUSH_FORMAT, maxdata=self._maxdata)
 
             with open(_local_path, 'rb') as stream:
@@ -628,15 +698,17 @@ class AdbDevice(object):
             The last modified time for the file
 
         """
+        if not device_path:
+            raise exceptions.DevicePathInvalidError("Cannot stat an empty device path")
         if not self.available:
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDevice.connect()`?)")
 
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+        adb_info = _AdbTransactionInfo(None, None, self._get_transport_timeout_s(transport_timeout_s), read_timeout_s)
         self._open(b'sync:', adb_info)
 
         filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_STAT_FORMAT, maxdata=self._maxdata)
         self._filesync_send(constants.STAT, adb_info, filesync_info, data=device_path)
-        _, (mode, size, mtime), _ = self._filesync_read([constants.STAT], adb_info, filesync_info, read_data=False)
+        _, (mode, size, mtime), _ = self._filesync_read([constants.STAT], adb_info, filesync_info)
         self._close(adb_info)
 
         return mode, size, mtime
@@ -709,6 +781,48 @@ class AdbDevice(object):
         if adb_info.local_id != their_local_id:
             raise exceptions.InvalidResponseError('Expected the local_id to be {}, got {}'.format(adb_info.local_id, their_local_id))
 
+    def _read_length(self, data_length, data_checksum, adb_info):
+        """Receive a response from the device.
+
+        1. Read a message's data from the device
+        2. If the checksum of the read data does not match ``data_checksum``, raise an exception
+
+
+        Parameters
+        ----------
+        data_length : int
+            We will read packets until we get this length of data
+        data_checksum: int
+            Data checksum
+        adb_info : _AdbTransactionInfo
+            Info and settings for this ADB transaction
+
+        Returns
+        -------
+        bytearray
+            The data that was read
+
+        Raises
+        ------
+        adb_shell.exceptions.InvalidChecksumError
+            Received checksum does not match the expected checksum.
+        """
+        data = bytearray()
+
+        if data_length > 0:
+            while data_length > 0:
+                temp = self._transport.bulk_read(data_length, adb_info.transport_timeout_s)
+                _LOGGER.debug("bulk_read(%d): %.1000r", data_length, temp)
+
+                data += temp
+                data_length -= len(temp)
+
+            actual_checksum = checksum(data)
+            if actual_checksum != data_checksum:
+                raise exceptions.InvalidChecksumError('Received checksum {0} != {1}'.format(actual_checksum, data_checksum))
+
+        return data
+
     def _read(self, expected_cmds, adb_info):
         """Receive a response from the device.
 
@@ -750,34 +864,19 @@ class AdbDevice(object):
 
         while True:
             msg = self._transport.bulk_read(constants.MESSAGE_SIZE, adb_info.transport_timeout_s)
-            _LOGGER.debug("bulk_read(%d): %s", constants.MESSAGE_SIZE, repr(msg))
+            _LOGGER.debug("bulk_read(%d): %r", constants.MESSAGE_SIZE, msg)
             cmd, arg0, arg1, data_length, data_checksum = unpack(msg)
             command = constants.WIRE_TO_ID.get(cmd)
 
             if not command:
                 raise exceptions.InvalidCommandError("Unknown command: %d = '%s' (arg0 = %d, arg1 = %d, msg = '%s')" % (cmd, int_to_cmd(cmd), arg0, arg1, msg))
 
+            data = self._read_length(data_length, data_checksum, adb_info)
             if command in expected_cmds:
                 break
 
             if time.time() - start > adb_info.read_timeout_s:
-                raise exceptions.InvalidCommandError("Never got one of the expected responses: %s (transport_timeout_s = %d, read_timeout_s = %d" % (expected_cmds, adb_info.transport_timeout_s, adb_info.read_timeout_s))
-
-        if data_length > 0:
-            data = bytearray()
-            while data_length > 0:
-                temp = self._transport.bulk_read(data_length, adb_info.transport_timeout_s)
-                _LOGGER.debug("bulk_read(%d): %.1000s", data_length, repr(temp))
-
-                data += temp
-                data_length -= len(temp)
-
-            actual_checksum = checksum(data)
-            if actual_checksum != data_checksum:
-                raise exceptions.InvalidChecksumError('Received checksum {0} != {1}'.format(actual_checksum, data_checksum))
-
-        else:
-            data = bytearray()
+                raise exceptions.InvalidCommandError("Never got one of the expected responses: %s (transport_timeout_s = %f, read_timeout_s = %f)" % (expected_cmds, adb_info.transport_timeout_s, adb_info.read_timeout_s))
 
         return command, arg0, arg1, bytes(data)
 
@@ -825,7 +924,7 @@ class AdbDevice(object):
                 break
 
             if time.time() - start > adb_info.read_timeout_s:
-                raise exceptions.InvalidCommandError("Never got one of the expected responses: %s (transport_timeout_s = %d, read_timeout_s = %d" % (expected_cmds, adb_info.transport_timeout_s, adb_info.read_timeout_s))
+                raise exceptions.InvalidCommandError("Never got one of the expected responses: %s (transport_timeout_s = %f, read_timeout_s = %f)" % (expected_cmds, adb_info.transport_timeout_s, adb_info.read_timeout_s))
 
             # Ignore CLSE responses to previous commands
             # https://github.com/JeffLIrion/adb_shell/pull/14
@@ -888,11 +987,12 @@ class AdbDevice(object):
             Info and settings for this ADB transaction
 
         """
-        _LOGGER.debug("bulk_write: %s", repr(msg.pack()))
-        self._transport.bulk_write(msg.pack(), adb_info.transport_timeout_s)
+        packed = msg.pack()
+        _LOGGER.debug("bulk_write(%d): %r", len(packed), packed)
+        self._transport.bulk_write(packed, adb_info.transport_timeout_s)
 
         if msg.data:
-            _LOGGER.debug("bulk_write: %s", repr(msg.data))
+            _LOGGER.debug("bulk_write(%d): %r", len(msg.data), msg.data)
             self._transport.bulk_write(msg.data, adb_info.transport_timeout_s)
 
     def _streaming_command(self, service, command, adb_info):
@@ -963,7 +1063,7 @@ class AdbDevice(object):
         self._write(filesync_info.send_buffer[:filesync_info.send_idx], adb_info)
         filesync_info.send_idx = 0
 
-    def _filesync_read(self, expected_ids, adb_info, filesync_info, read_data=True):
+    def _filesync_read(self, expected_ids, adb_info, filesync_info):
         """Read ADB messages and return FileSync packets.
 
         Parameters
@@ -974,8 +1074,6 @@ class AdbDevice(object):
             Info and settings for this ADB transaction
         filesync_info : _FileSyncTransactionInfo
             Data and storage for this FileSync transaction
-        read_data : bool
-            Whether to read the received data
 
         Returns
         -------
@@ -984,7 +1082,7 @@ class AdbDevice(object):
         tuple
             The contents of the header
         data : bytearray, None
-            The received data, or ``None`` if ``read_data`` is False
+            The received data, or ``None`` if the command ID is :const:`adb_shell.constants.STAT`
 
         Raises
         ------
@@ -1000,14 +1098,23 @@ class AdbDevice(object):
         # Read one filesync packet off the recv buffer.
         header_data = self._filesync_read_buffered(filesync_info.recv_message_size, adb_info, filesync_info)
         header = struct.unpack(filesync_info.recv_message_format, header_data)
+
         # Header is (ID, ...).
         command_id = constants.FILESYNC_WIRE_TO_ID[header[0]]
 
+        # Whether there is data to read
+        read_data = command_id != constants.STAT
+
+        if read_data:
+            # Header is (ID, ..., size) --> read the data
+            data = self._filesync_read_buffered(header[-1], adb_info, filesync_info)
+        else:
+            # No data to be read
+            data = bytearray()
+
         if command_id not in expected_ids:
             if command_id == constants.FAIL:
-                reason = ''
-                if filesync_info.recv_buffer:
-                    reason = filesync_info.recv_buffer.decode('utf-8', errors='ignore')
+                reason = data.decode('utf-8', errors='ignore')
 
                 raise exceptions.AdbCommandFailureException('Command failed: {}'.format(reason))
 
@@ -1016,10 +1123,6 @@ class AdbDevice(object):
         if not read_data:
             return command_id, header[1:], None
 
-        # Header is (ID, ..., size).
-        size = header[-1]
-        data = self._filesync_read_buffered(size, adb_info, filesync_info)
-
         return command_id, header[1:-1], data
 
     def _filesync_read_buffered(self, size, adb_info, filesync_info):
@@ -1082,7 +1185,7 @@ class AdbDevice(object):
             if cmd_id in finish_ids:  # pragma: no cover
                 break
 
-    def _filesync_send(self, command_id, adb_info, filesync_info, data=b'', size=0):
+    def _filesync_send(self, command_id, adb_info, filesync_info, data=b'', size=None):
         """Send/buffer FileSync packets.
 
         Packets are buffered and only flushed when this connection is read from. All
@@ -1098,13 +1201,13 @@ class AdbDevice(object):
             Data and storage for this FileSync transaction
         data : str, bytes
             Optional data to send, must set data or size.
-        size : int
+        size : int, None
             Optionally override size from len(data).
 
         """
-        if data:
-            if not isinstance(data, bytes):
-                data = data.encode('utf8')
+        if not isinstance(data, bytes):
+            data = data.encode('utf8')
+        if size is None:
             size = len(data)
 
         if not filesync_info.can_add_to_send_buffer(len(data)):
@@ -1154,14 +1257,18 @@ class AdbDeviceTcp(AdbDevice):
         Whether an ADB connection to the device has been established
     _banner : bytearray, bytes
         The hostname of the machine where the Python interpreter is currently running
+    _default_transport_timeout_s : float, None
+        Default timeout in seconds for TCP packets, or ``None``
+    _maxdata : int
+        Maximum amount of data in an ADB packet
     _transport : TcpTransport
         The transport that is used to connect to the device
 
     """
 
     def __init__(self, host, port=5555, default_transport_timeout_s=None, banner=None):
-        transport = TcpTransport(host, port, default_transport_timeout_s)
-        super(AdbDeviceTcp, self).__init__(transport, banner)
+        transport = TcpTransport(host, port)
+        super(AdbDeviceTcp, self).__init__(transport, default_transport_timeout_s, banner)
 
 
 class AdbDeviceUsb(AdbDevice):
@@ -1190,6 +1297,10 @@ class AdbDeviceUsb(AdbDevice):
         Whether an ADB connection to the device has been established
     _banner : bytearray, bytes
         The hostname of the machine where the Python interpreter is currently running
+    _default_transport_timeout_s : float, None
+        Default timeout in seconds for USB packets, or ``None``
+    _maxdata : int
+        Maximum amount of data in an ADB packet
     _transport : UsbTransport
         The transport that is used to connect to the device
 
@@ -1200,4 +1311,4 @@ class AdbDeviceUsb(AdbDevice):
             raise exceptions.InvalidTransportError("To enable USB support you must install this package via `pip install adb-shell[usb]`")
 
         transport = UsbTransport.find_adb(serial, port_path, default_transport_timeout_s)
-        super(AdbDeviceUsb, self).__init__(transport, banner)
+        super(AdbDeviceUsb, self).__init__(transport, default_transport_timeout_s, banner)
